from dash_labs.templates.base import BaseTemplate
import dash_html_components as html


class HtmlCard(BaseTemplate):
    """
    Simple template that places all components in a few html Div elements with a
    card-like border.
    """

    def __init__(self, title=None, width=None):
        super().__init__()
        self.title = title
        self.width = width

    def _perform_layout(self):
        # No callbacks here. Must be constant or idempotent
        children = []
        if self.title:
            children.append(html.H2(self.title))

        children.append(html.Div(self.get_containers("output")))
        children.append(html.Hr())
        children.append(html.Div(self.get_containers("input")))
        layout = html.Div(
            style={
                "width": self.width,
                "border": "1px solid lightgray",
                "padding": 10,
                "border-radius": "6px",
            },
            children=html.Div(children=children),
        )
        return layout



from ._callback import _callback as dx_callback
from functools import partial
from dash import Dash
from types import MethodType


class FlexibleCallbacks:
    """
    Dash app plugin to enable advanced callback behaviors including:
      - Property grouping
      - Keyword argument support
      - Support for providing full components in place of ids when using the
        dl.Output/dl.Input/sl.State dependency objects.

    Usage:

    >>> import dash
    >>> import dash_labs as dl
    >>> app = dash.Dash(__name__, plugins=[dl.FlexibleCallbacks()])
    """

    def __init__(self):
        pass

    def plug(self, app):
        _wrapped_callback = Dash.callback
        app._wrapped_callback = _wrapped_callback
        app.callback = MethodType(
            partial(dx_callback, _wrapped_callback=_wrapped_callback), app
        )



import math

from dash_labs.dependency import Output, Input
from dash_labs.util import build_id, filter_kwargs
from .base import ComponentPlugin
from dash_labs.templates import FlatDiv


operators = [
    ["ge ", ">="],
    ["le ", "<="],
    ["lt ", "<"],
    ["gt ", ">"],
    ["ne ", "!="],
    ["eq ", "="],
    ["contains "],
    ["datestartswith "],
]


class DataTablePlugin(ComponentPlugin):
    """
    Component plugin to make it easier to work with DataTables
    """

    def __init__(
        self,
        df,
        columns=None,
        page_size=5,
        sort_mode=None,
        filterable=False,
        serverside=False,
        template=None,
        role="output",
    ):
        """

        :param df: Pandas DataFrame to be displayed in the DataTable
        :param columns: List of columns from df to display in the table. Defaults to
            all columns if not specified.
        :param page_size: The number of rows of df to display at once. If the number
            of rows exceeds this number, the table will display paging controls.
        :param sort_mode: Sort mode for table. One of:
            - None: no sorting
            - 'single': single column sorting
            - 'multi': multi-column sorting
        :param filterable: If True, display a filter box below each column in the
            table. If False (default), no filtering interface is displayed.
        :param serverside: If True, paging/sorting/filtering operations trigger
            a callback and are performed in Python and only the current displayed
            page is transferred to the client. If False (default), all rows of the table
            are sent to the client at startup and paging/sorting/filtering
            operations do not trigger a Python callback.
        :param template: Template to use to construct DataTable
        :param role: Template role that should be assigned to the constructed DataTable
        """
        if template is None:
            template = FlatDiv()

        if columns is None:
            columns = list(df.columns)

        self.page_size = page_size
        self.sort_mode = sort_mode
        self.filterable = filterable
        self.role = role

        self.serverside = serverside
        self.page_count = self._compute_page_count(df)

        if self.serverside:
            self.full_df = df
            self.df = self._compute_serverside_dataframe_slice(df)
        else:
            self.full_df = df
            self.df = df

        self.data, self.columns = self.convert_data_columns(self.df, columns)
        self.datatable_id = build_id()

        # Initialize args
        if self.serverside:
            args = self._build_serverside_input(template)
            output = self._build_serverside_output(template)
        else:
            args = self._build_clientside_input(template)
            output = self._build_clientside_output(template)

        super().__init__(args, output, template)

    def get_output_values(self, args_value, df=None, preprocessed=False):
        """
        :param args_value: Callback arguments corresponding to the args dependenc
            grouping
        :param df: DataFrame to use as input for the table
            (prior to paging/filtering/sorting). If not provided, the DataFrame passed
            to the constructor is used.
        :param preprocessed: Set to true if the df argument was produced by the
            get_processed_dataframe method. This will bypass the serverside filtering
            and sorting logic, since it was already applied by get_processed_dataframe
        :return: Grouping of values corresponding to the dependency grouping returned
            by the output property
        """
        if self.serverside:
            return self._build_serverside_result(
                args_value, df, preprocessed=preprocessed
            )
        else:
            return self._build_clientside_result(df)

    def get_processed_dataframe(self, args_value, df=None):
        """
        Retrieve the DataFrame produced by the serverside filtering and sorting
        operations.

        Note: This method is only supported when serverside=True

        :param args_value: Callback arguments corresponding to the args dependenc
            grouping
        :param df: DataFrame to use as input to the table filtering and sorting
            operations. If not provided, the DataFrame passed to the constructor is
            used.
        :return: DataFrame that has
        """
        if not self.serverside:
            raise ValueError(
                "get_processed_dataframe is only supported when serverside=True"
            )

        sort_by = args_value["sort_by"]
        # Get active dataframe
        if df is None:
            df = self.full_df
        # Perform filtering
        if self.filterable and "filter_query" in args_value:
            filter_query = args_value["filter_query"]
            df = _filter_serverside(df, filter_query)
        # Perform sorting
        if sort_by and len(sort_by):
            df = df.sort_values(
                [col["column_id"] for col in sort_by],
                ascending=[col["direction"] == "asc" for col in sort_by],
            )
        return df

    def _compute_serverside_dataframe_slice(self, full_df, page_current=None):
        if page_current is None:
            page_current = 0

        start_ind = page_current * self.page_size
        end_ind = start_ind + self.page_size
        return full_df.iloc[start_ind:end_ind]

    def _compute_page_count(self, full_df):
        return math.ceil(len(full_df) / self.page_size)

    def convert_data_columns(self, df, columns=None):
        # Handle DataFrame input
        import pandas as pd

        if isinstance(df, pd.DataFrame):
            if columns is None:
                columns = df.columns.tolist()
            df = df.to_dict("records")

        # Handle columns as list
        if isinstance(columns, list) and columns and not isinstance(columns[0], dict):
            columns = [{"name": col, "id": col} for col in columns]

        return df, columns

    # Serverside helpers
    def _build_serverside_input(self, template):
        return {
            "page_current": Input(self.datatable_id, "page_current"),
            "sort_by": Input(self.datatable_id, "sort_by"),
            "filter_query": Input(self.datatable_id, "filter_query"),
        }

    def _build_serverside_output(self, template):
        data, columns = self.convert_data_columns(self.df, self.columns)
        result = Output(
            template._datatable_class()(
                data=data,
                columns=columns,
                id=self.datatable_id,
                page_current=0,
                page_size=self.page_size,
                page_action="custom",
                page_count=self._compute_page_count(self.full_df),
                **filter_kwargs(
                    sort_action=None if self.sort_mode is None else "custom",
                    sort_mode=self.sort_mode,
                    filter_action="custom" if self.filterable else None,
                    filter_query="" if self.filterable else None,
                )
            ),
            component_property=dict(
                data="data",
                columns="columns",
                page_count="page_count",
            ),
            role=self.role,
        )
        return result

    def _build_serverside_result(self, inputs_value, df, preprocessed=False):
        page_current = inputs_value["page_current"]

        if not preprocessed:
            df = self.get_processed_dataframe(inputs_value, df)

        df_slice = self._compute_serverside_dataframe_slice(
            df, page_current=page_current
        )

        data, columns = self.convert_data_columns(df_slice)
        page_count = self._compute_page_count(df)
        return dict(data=data, columns=columns, page_count=page_count)

    # Clientside helpers
    def _build_clientside_input(self, template):
        return ()

    def _build_clientside_output(self, template):
        data, columns = self.convert_data_columns(self.df, self.columns)
        return Output(
            template._datatable_class()(
                data=data,
                columns=columns,
                id=self.datatable_id,
                page_size=self.page_size,
                **filter_kwargs(
                    sort_action=None if self.sort_mode is None else "native",
                    sort_mode=self.sort_mode,
                    filter_action="native" if self.filterable else None,
                )
            ),
            component_property=dict(data="data", columns="columns"),
            role=self.role,
        )

    def _build_clientside_result(self, df):
        if df is not None:
            data, columns = self.convert_data_columns(df)
        else:
            data, columns = self.data, self.columns

        return dict(data=data, columns=columns)


def _split_filter_part(filter_part):
    for operator_type in operators:
        for operator in operator_type:
            if operator in filter_part:
                name_part, value_part = filter_part.split(operator, 1)
                name = name_part[name_part.find("{") + 1 : name_part.rfind("}")]

                value_part = value_part.strip()
                v0 = value_part[0]
                if v0 == value_part[-1] and v0 in ("'", '"', "`"):
                    value = value_part[1:-1].replace("\\" + v0, v0)
                else:
                    try:
                        value = float(value_part)
                    except ValueError:
                        value = value_part

                # word operators need spaces after them in the filter string,
                # but we don't want these later
                return name, operator_type[0].strip(), value

    return [None] * 3


def _filter_serverside(df, filter_query):
    filtering_expressions = filter_query.split(" && ")
    dff = df

    for filter_part in filtering_expressions:
        col_name, operator, filter_value = _split_filter_part(filter_part)

        if operator in ("eq", "ne", "lt", "le", "gt", "ge"):
            # these operators match pandas series operator method names
            dff = dff.loc[getattr(dff[col_name], operator)(filter_value)]
        elif operator == "contains":
            dff = dff.loc[dff[col_name].str.contains(filter_value)]
        elif operator == "datestartswith":
            # this is a simplification of the front-end filtering logic,
            # only works with complete fields in standard format
            dff = dff.loc[dff[col_name].str.startswith(filter_value)]

    return dff



from dash.dependencies import (
    Input as Input_dash,
    Output as Output_dash,
    State as State_dash,
)
from dash.development.base_component import Component

from dash_labs import build_id
from dash_labs.grouping import make_grouping_by_index, flatten_grouping, map_grouping

"""
Module containing DashLabs versions of the standard Input/State/Output dash
dependencies. These versions support the use of full components in place of component
ids, and they support specifying property groupings in place of scalar properties.
"""


class DashLabsDependency:

    dependency_class = None

    def __init__(
        self,
        component_id,
        component_property="value",
        label=Component.UNDEFINED,
        role=Component.UNDEFINED,
        label_id=None,
    ):
        """
        :param component_id: Component id or component instance
        :param component_property: Grouping of component properties
        :param label: Template label for component
            (ignored if component_id is not a component)
        :param role: Template role for component
            (ignored if component_id is not a component)
        :param label_id: Custom id to use for template label
            (ignored if component_id is not a component or label is not defined)
        """
        self.set_component_and_props(component_id, component_property)
        self.label = label
        self.label_id = label_id
        self.role = role

    @property
    def id(self):
        """
        :return: Component's id if component_id is a component, otherwise component_id
        """
        if isinstance(self.component_id, (str, dict)):
            return self.component_id
        else:
            return self.component_id.id

    @property
    def has_component(self):
        """
        :return: Trued if component_id is a component instance, False otherwise
        """
        return isinstance(self.component_id, Component)

    def extract_component(self):
        """
        Extract the store component from dependency and construct dependency that
        references component by id

        :return: (dependency, component) tuple
        """
        if not self.has_component:
            raise ValueError(
                "extract_component only valid for dependency storing a component instance"
            )
        component = self.component_id
        dependency = type(self)(self.id, self.component_property)
        return component, dependency

    def set_component_and_props(self, component_id, component_property):
        """
        Set both component_id and component_property attributes

        :param component_id: New component_id or component instance
        :param component_property:  New component property grouping
        """
        if isinstance(component_id, Component):
            _validate_prop_grouping(component_id, component_property)
            self.component_id = component_id
            self.component_property = component_property

            if getattr(self.component_id, "id", None) is None:
                self.component_id.id = build_id()
        else:
            if not isinstance(component_id, (str, dict)):
                raise ValueError("Invalid component_id value: {}".format(component_id))
            self.component_id = component_id
            self.component_property = component_property

    def dependencies(self, labs=False):
        """
        :param labs: If True, return dash-labs dependency objects
            if False (default), return dash.dependency objects
        :return: Grouping of dash.dependency objects with structure matching the
            grouping structore of the component_property attribute
        """
        if labs:
            return self._make_dependency_grouping(self.__class__)
        else:
            return self._make_dependency_grouping(self.dependency_class)

    def flat_dependencies(self, labs=False):
        """
        :param labs: If True, return dash-labs dependency objects
            if False (default), return dash.dependency objects
        :return: Flat list of dash.dependency objects with length matching the number
            of scalar properties in the component_property grouping
        """
        if labs:
            return self._make_flat_dependencies(self.__class__)
        else:
            return self._make_flat_dependencies(self.dependency_class)

    @property
    def flat_props(self):
        """
        :return: Flat list of properties in component_property grouping
        """
        return flatten_grouping(self.component_property)

    def _make_flat_dependencies(self, dependency):
        return [dependency(self.id, prop) for prop in self.flat_props]

    def _make_dependency_grouping(self, dependency):
        return make_grouping_by_index(
            self.component_property, self._make_flat_dependencies(dependency)
        )

    def property_value(self):
        """
        :return: Return a grouping of component property values corresponding to the
            component_property grouping
        """
        if not self.has_component:
            raise ValueError(
                "property_value only valid for dependency storing a component instance"
            )
        return map_grouping(
            lambda p: getattr(self.component_id, p, None), self.component_property
        )

    def __str__(self):
        return (
            f"{self.__class__.__name__}(\n"
            f"    component_id={repr(self.component_id)},\n"
            f"    component_property={repr(self.component_property)}\n"
            f")"
        )

    def __repr__(self):
        return str(self)


class Input(DashLabsDependency):
    dependency_class = Input_dash


class State(DashLabsDependency):
    dependency_class = State_dash


class Output(DashLabsDependency):
    dependency_class = Output_dash


def _validate_prop_grouping(component, props):
    for prop in flatten_grouping(props):
        _validate_prop_name(component, prop)


def _validate_prop_name(component, prop_name):
    if (
        isinstance(prop_name, str)
        and prop_name.isidentifier()
        and prop_name in component._prop_names
    ):
        pass
    else:
        raise ValueError(
            "Invalid property {prop} received for component of type {typ}\n".format(
                prop=repr(prop_name), typ=type(component)
            )
        )



from collections import OrderedDict
from dash_labs.dependency import Output, Input
from dash_labs.templates.base import BaseTemplate
import dash_html_components as html
from dash_labs.util import filter_kwargs, build_id


class BaseDDKTemplate(BaseTemplate):
    """
    Base class for templates based on Dash Design Kit
    """

    _label_value_prop = "label"

    def __init__(
        self,
        theme=None,
        **kwargs,
    ):
        """
        :param theme: DDK theme dictionary
        :param kwargs: Additional keyword arguments to pass to ddk.App
        """
        super(BaseDDKTemplate, self).__init__()
        self.theme = theme
        self.app_kwargs = kwargs

    @classmethod
    def build_labeled_component(cls, component, label, label_id=None, role=None):
        import dash_design_kit as ddk

        # Subclass could use bootstrap or ddk
        if not label_id:
            label_id = build_id("label")

        container = ddk.ControlItem(id=label_id, label=label, children=component)
        label_component = container
        label_property = "label"

        return container, "children", label_component, label_property

    @classmethod
    def build_containered_component(cls, component, role=None):
        import dash_design_kit as ddk

        # Subclass could use bootstrap or ddk
        container_id = build_id("container")
        container = ddk.ControlItem(id=container_id, children=component)

        return container, "children"

    @classmethod
    def _graph_class(cls):
        import dash_design_kit as ddk

        return ddk.Graph

    @classmethod
    def _datatable_class(cls):
        from dash_table import DataTable

        return DataTable

    def _wrap_full_layout(self, layout):
        import dash_design_kit as ddk

        return ddk.App(
            children=layout,
            **filter_kwargs(self.app_kwargs, theme=self.theme),
        )


class DdkCard(BaseDDKTemplate):
    def __init__(self, title=None, width=None, height=None, **kwargs):
        """
        Template that places all components in a single card

        :param title: Card title
        :param width: Proportional width of the card (out of 100)
        :param height: Height of the card (in pixels)
        """
        super().__init__(**kwargs)
        self.title = title
        self.width = width
        self.height = height

    def _perform_layout(self):
        import dash_design_kit as ddk

        # No callbacks here. Must be constant or idempotent
        card_children = []
        if self.title:
            card_children.append(ddk.CardHeader(title=self.title))

        card_children.append(
            html.Div(
                style={"padding": 20}, children=html.Div(self.get_containers("output"))
            )
        )
        card_children.append(html.Hr(style={"width": "100%", "margin": "auto"}))
        card_children.extend(self.get_containers("input"))

        layout = ddk.ControlCard(
            children=card_children,
            **filter_kwargs(width=self.width),
        )

        return layout


class DdkRow(BaseDDKTemplate):
    def __init__(self, title=None, input_width=30, **kwargs):
        """
        Template that places inputs and outputs in separate cards, arranged in a row

        :param title: Input card title
        :param input_width: Input width proportion (out of 100)
        """
        super().__init__(**kwargs)
        self.title = title
        self.input_width = input_width

    def _perform_layout(self):
        import dash_design_kit as ddk

        # Input card
        input_card = ddk.ControlCard(
            children=self.get_containers("input"),
            width=self.input_width,
        )

        output_card_children = []
        if self.title is not None:
            output_card_children.append(ddk.CardHeader(title=self.title))
        output_card_children.extend(self.get_containers("output"))

        output_card = ddk.ControlCard(
            children=output_card_children,
            width=100 - self.input_width,
        )

        row_children = [input_card, output_card]
        layout = ddk.Row(row_children)

        return layout


class DdkSidebar(BaseDDKTemplate):
    def __init__(self, title=None, sidebar_width=300, **kwargs):
        """
        Template that includes a title bar, places inputs in a sidebar, and outputs in
        the main area of the app.

        :param title: Title bar title string
        :param sidebar_width: Sidebar width in pixels or as a css string
        """
        super().__init__(**kwargs)
        self.title = title
        self.sidebar_width = sidebar_width

    def _perform_layout(self):
        import dash_design_kit as ddk

        children = []

        if self.title is not None:
            children.append(ddk.Header([ddk.Title(self.title)]))

        # Input card
        sidebar_children = []
        sidebar_children.append(
            ddk.ControlCard(
                children=self.get_containers("input"),
            )
        )

        sidebar = ddk.Sidebar(
            foldable=True,
            children=sidebar_children,
            style={"minWidth": self.sidebar_width if self.sidebar_width else "auto"},
        )
        children.append(sidebar)

        output_card_children = []
        output_card_children.extend(self.get_containers("output"))

        output_card = ddk.ControlCard(output_card_children)

        sidebar_companion = ddk.SidebarCompanion(output_card)
        children.append(sidebar_companion)

        return children


class DdkSidebarTabs(BaseDDKTemplate):
    def __init__(self, tab_roles, title=None, sidebar_width=300, **kwargs):
        """
        Template that includes a title bar, places inputs in a sidebar, and outputs in
        a set of tabs

        :param tab_roles: List or dict of strings where each string specifies the name
            of the role corresponding to a single tab. If a list, the role name is
            also be used as the title of the corresponding tab. If a dict, the keys
            become the roles and the values become the tab labels
        :param title: Title bar title string
        :param sidebar_width: Sidebar width in pixels or as a css string
        """
        import dash_core_components as dcc

        self.title = title
        self.sidebar_width = sidebar_width
        if isinstance(tab_roles, (list, tuple)):
            self.tab_roles = OrderedDict([(role, role) for role in tab_roles])
        else:
            self.tab_roles = OrderedDict(tab_roles)

        self._valid_roles = ["input", "output"] + list(self.tab_roles.keys())
        self._tabs = dcc.Tabs(id=build_id("tabs"), value=self._valid_roles[2])

        super().__init__(**kwargs)

    def _perform_layout(self):
        import dash_design_kit as ddk
        import dash_core_components as dcc

        children = []

        if self.title is not None:
            children.append(ddk.Header([ddk.Title(self.title)]))

        # Input card
        sidebar_children = []
        sidebar_children.append(
            ddk.ControlCard(
                children=self.get_containers("input"),
            )
        )

        sidebar = ddk.Sidebar(
            foldable=True,
            children=sidebar_children,
            style={"minWidth": self.sidebar_width if self.sidebar_width else "auto"},
        )
        children.append(sidebar)

        self._tabs.children = [
            dcc.Tab(
                value=role,
                label=title,
                children=ddk.ControlCard(list(reversed(self.get_containers(role)))),
            )
            for role, title in self.tab_roles.items()
        ]

        sidebar_companion = ddk.SidebarCompanion(self._tabs)
        children.append(sidebar_companion)

        return children

    def tab_input(self, kind=Input):
        """
        Dependency object that can be used to input the active tab
        :param kind: The dependency kind to return. One of dl.Input (default) or
            dl.State.
        :return: Dependency object referencing the active tab
        """
        return kind(self._tabs.id, "value")



import sys
from typing import Union

import dash_html_components as html
import dash_core_components as dcc
import dash_table  # noqa: Needs table initialization
from collections import OrderedDict
import datetime

from dash.development.base_component import Component

from dash_labs.util import filter_kwargs, build_id, insert_into_ordered_dict
from dash_labs.dependency import Input, Output, State, DashLabsDependency
from dataclasses import dataclass


@dataclass(frozen=True)
class ArgumentComponents:
    arg_component: Component
    arg_property: Union[str, tuple, dict]
    label_component: Component
    label_property: Union[str, tuple, dict]
    container_component: Component
    container_property: Union[str, tuple, dict]


class BaseTemplate:
    """
    Base class for dash-labs templates
    """

    # The property of this template's label components that holds the label string
    _label_value_prop = "children"

    # Optional string containing css class definitions that will be enclosed in a
    # <style></style> tag and added to the app's index.html file.
    _inline_css = """
    .rc-slider-tooltip-inner {
        user-select: none;
    }
    """

    # Tuple of the roles supported by this template. Subclasses may override this
    # as a class attribute (as is the case here), or as an instance attribute if the
    # available roles are dependent on constructor arguments.
    #
    # If overriding as an instance attribute, be sure to set the value of the
    # _valid_roles attribute before calling the superclass constructor so that the
    # self._roles dict is initialized properly
    _valid_roles = ("input", "output")

    def __init__(self):
        self._roles = {role: OrderedDict() for role in self._valid_roles}

    @property
    def roles(self):
        """
        Dictionary from role to OrderedDict of ArgumentComponents instances.

        Each component added to the template is wrapped in an ArgumentComponents
        instance and stored in the OrderedDict corresponding to the component's role

        :return: dict from roles to OrderedDict of ArgumentComponents instances
        """
        return self._roles

    @classmethod
    def build_argument_components(
        cls,
        component,
        value_property=(),
        label=None,
        label_id=None,
        role=None,
    ):
        # Get reference to dependency class object for role
        arg_component = component
        arg_props = value_property

        if label and label is not Component.UNDEFINED:
            initial_value = label
            (
                container_component,
                container_props,
                label,
                label_props,
            ) = cls.build_labeled_component(
                arg_component,
                label=initial_value,
                label_id=label_id,
                role=role,
            )
            label_component = label
            label_props = label_props
        else:
            label_component = None
            label_props = None
            container_component, container_props = cls.build_containered_component(
                arg_component,
                role=role,
            )

        return ArgumentComponents(
            arg_component=arg_component,
            arg_property=arg_props,
            label_component=label_component,
            label_property=label_props,
            container_component=container_component,
            container_property=container_props,
        )

    def add_component(
        self,
        component,
        role="input",
        label=None,
        label_id=None,
        name=None,
        component_property=None,
        before=None,
        after=None,
    ):
        """
        Add a component to the template

        :param component: The Dash component instance to add to the template
        :param role: The role of the component within the template. All templates
            support the "input" and "output" roles, but individual templates may
            support additional roles.
        :param label: (optional) A string label for the component
        :param label_id: (optional) A custom component id to use for the created
            label component (if any)
        :param name: (optional) A string name for the component. Must be unique for
            all components within the same role. If not provided, name is effectively
            the positional index of component within role.
        :param component_property: (optional) A component property to be stored in the
            ArgumentComponents instance for the added component
        :param before: (optional) String name, or positional index, of existing
            component within the same role that this component should be inserted
            before.
        :param after: (optional) String name, or positional index, of existing
            component within the same role that this component should be inserted
            after.
        :return: ArgumentComponents instance representing the component and label
            within the template
        """
        if role not in self._valid_roles:
            raise ValueError(
                "Invalid role {role} provided for template of type {typ}\n"
                "    Supported roles: {roles}".format(
                    role=repr(role), typ=type(self), roles=self._valid_roles
                )
            )

        arg_components = self.build_argument_components(
            component,
            value_property=component_property,
            label=label,
            label_id=label_id,
            role=role,
        )

        self._roles[role] = insert_into_ordered_dict(
            odict=self._roles[role],
            value=arg_components,
            key=name,
            before=before,
            after=after,
        )
        return arg_components

    def get_containers(self, roles=None):
        """
        Get list of Dash components that contain the components added to the template

        :param roles: String name of single role to include, or list of
            roles to include. If None (default), include components from all roles
        :return: flat list of container components
        """
        if roles is None:
            roles = self._valid_roles
        elif isinstance(roles, str):
            roles = [roles]

        containers = []
        for role in roles:
            for pc in self._roles.get(role, {}).values():
                containers.append(pc.container_component)

        return containers

    def layout(self, app, full=True):
        """
        Build and return a component that aranges all of the components added to the
        template.

        May also configure application with additional css and scripts required by
        the template.

        :param app: The dash.Dash app instance that will be used to display the
            template
        :param full: Some component toolkits require a top-level container component
            that should be assigned to app.layout. For example, Dash Bootstrap
            Components requires `dbc.Container` and Dash Design Kit requires ddk.App.
             - If full=True (the default) the returned layout is wrapped in this
               top-level component and so it may be directly assigned to app.layout.
               In this case, the returned template should not be nested inside of a
               larger app.
             - If full=False, the returned layout is not wrapped in this top-level
               component. This makes it possible to embed the returned layout in a
               larger app, but it is the caller's responsibility to wrap the full app
               in the appropriate top-level component.

        :return: Dash Component
        """
        # Build structure
        layout = self._perform_layout()
        if full:
            layout = self._wrap_full_layout(layout)

        # Configure app props like CSS
        self._configure_app(app)

        return layout

    # Methods below this comment are designed to be customized by template subclasses
    # -------------------------------------------------------------------------------
    def _perform_layout(self):
        """
        Build the layout, with the exception of any special top-level components
        added by `_wrap_layout`.

        :return: Dash Component
        """
        raise NotImplementedError

    @classmethod
    def _wrap_full_layout(cls, layout):
        """
        Wrap layout in a top-level component (e.g. Ddk.App or dbc.Container)
        :param layout: layout component
        :return: Top-level Dash Component
        """
        return layout

    def _configure_app(self, app):
        """
        Configure application object.

        e.g. install inline CSS / add external stylsheets
        """
        if self._inline_css and self._inline_css not in app.index_string:
            new_css = "\n<style>{}</style>\n".format(self._inline_css)
            app.index_string = app.index_string.replace("{%css%}", "{%css%}" + new_css)

    @classmethod
    def build_labeled_component(cls, component, label, label_id=None, role=None):
        """
        Wrap input component in a labeled container

        :param component: Component to wrap
        :param label: Label string
        :param label_id: (optional) component of component that will store the label
            string
        :param role: Component role
        :return: tuple of (
            container: Dash component containing input component and label component
            container_property: Property of container component containing the input
                component
            label_component: Dash component containing the label string
            label_property: Property of label_component that stores the label string
        )
        """
        # Subclass could use bootstrap or ddk
        if not label_id:
            label_id = build_id("label")
        label_component = html.Label(id=label_id, children=label)
        container_id = build_id("container")
        container = html.Div(
            id=container_id,
            style={"padding-top": 10},
            children=[
                label_component,
                html.Div(style={"display": "block"}, children=component),
            ],
        )
        return container, "children", label_component, "children"

    @classmethod
    def build_containered_component(cls, component, role=None):
        """
        Alternative to bulid_labeled_component for use without label. Used to
        provided unitform spacing across labeled and unlabeled components

        :param component: Component to wrap in container
        :param role: Component role
        :return: tuple of (
            container: Dash component containing input component
            container_property: Property of container component containing the input
                component
        )
        """
        container_id = build_id("container")
        container = html.Div(
            id=container_id,
            style={"padding-top": 10},
            children=[
                html.Div(style={"display": "block"}, children=component),
            ],
        )
        return container, "children"

    @classmethod
    def default_output(cls, children=None):
        """
        If not Output dependency is specified in app.callback, use this component
        dependency as the output.

        Typically an html.Div, but it may be overridden by template subclasses

        :return: Dash Component
        """
        return cls.div_output(**filter_kwargs(children=children))

    # Component dependency constructors
    # ---------------------------------
    # Subclasses are welcome to override component dependency constructors, and to add
    # new constructors. The following conventions should be followed
    #
    #  1. The final underscore suffix should match the default role of the component.
    #     e.g. all of the examples below have either a _input or _output suffix, which
    #     matches the default value of the role argument.  The default suffix/role
    #     should be chosen based on whether the component is most commonly used as an
    #     input or output to a callback function.  Callers can override this behavior
    #     using the role and kind arguments.
    #
    #     If there are common uses for the component type in multiple roles,
    #     multiple methods may be defined, with a suffix for each role.
    #
    #  2. The default value of the kind argument should be consistent with the default
    #     role (e.g. kind=dl.Input for role="input" and kind=hl.Output for
    #     role="output"). The return value of the method should always match the class
    #     passed as the kind argument.
    #
    #  3. The default value of component_property should be a grouping of properties
    #     from the type of component returned that are most likely to be used as
    #     the input or output properties of the callback for that component.
    #
    #  4  The opts dict argument should be passed through as keyword arguments to the
    #     constructor of the underlying component.
    #
    #  5. The optional argument should be used to override the id of the constructed
    #     component.
    #
    #  6. The component_property, label, and role arguments should be passed through
    #     to the return dependency object.
    #
    @classmethod
    def div_output(
        cls,
        children=None,
        label=Component.UNDEFINED,
        role="output",
        component_property="children",
        kind=Output,
        id=None,
        opts=None,
    ):
        return kind(
            html.Div(**filter_kwargs(opts, children=children, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def markdown_output(
        cls,
        children=None,
        label=Component.UNDEFINED,
        role="output",
        component_property="children",
        kind=Output,
        id=None,
        opts=None,
    ):
        return kind(
            dcc.Markdown(**filter_kwargs(opts, children=children, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def textarea_input(
        cls,
        value=None,
        label=Component.UNDEFINED,
        role="input",
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        return kind(
            dcc.Textarea(**filter_kwargs(opts, value=value, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def button_input(
        cls,
        children,
        label=Component.UNDEFINED,
        role="input",
        component_property="n_clicks",
        kind=Input,
        id=None,
        opts=None,
    ):
        return kind(
            html.Button(children=children, **filter_kwargs(opts, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def dropdown_input(
        cls,
        options,
        value=Component.UNDEFINED,
        clearable=False,
        label=Component.UNDEFINED,
        role="input",
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        if isinstance(options, list) and options and not isinstance(options[0], dict):
            options = [{"value": opt, "label": opt} for opt in options]

        # Set starting value if dropdown is not clearable
        if value is Component.UNDEFINED and options and not clearable:
            value = options[0]["value"]

        return kind(
            dcc.Dropdown(
                options=options,
                **filter_kwargs(opts, value=value, id=id, clearable=clearable)
            ),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def slider_input(
        cls,
        min,
        max,
        value=Component.UNDEFINED,
        step=None,
        tooltip=None,
        label=Component.UNDEFINED,
        role="input",
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        if tooltip is None:
            tooltip = (opts or {}).pop(
                "tooltip", {"placement": "bottom", "always_visible": True}
            )
        elif tooltip is True:
            tooltip = {"placement": "bottom", "always_visible": True}
        else:
            tooltip = None

        if value is Component.UNDEFINED:
            value = min

        return kind(
            dcc.Slider(
                min=min,
                max=max,
                **filter_kwargs(opts, tooltip=tooltip, step=step, value=value, id=id)
            ),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def textbox_input(
        cls,
        value=None,
        label=Component.UNDEFINED,
        role="input",
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        return kind(
            dcc.Input(value=value, **filter_kwargs(opts, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def checklist_input(
        cls,
        options,
        value=None,
        label=Component.UNDEFINED,
        role="input",
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        if isinstance(options, list) and options and not isinstance(options[0], dict):
            options = [{"value": opt, "label": opt} for opt in options]

        if not isinstance(value, list):
            value = [value]

        return kind(
            dcc.Checklist(options=options, **filter_kwargs(opts, value=value, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def _graph_class(cls):
        return dcc.Graph

    @classmethod
    def graph_output(
        cls,
        figure=None,
        config=None,
        label=Component.UNDEFINED,
        role="output",
        component_property="figure",
        kind=Output,
        id=None,
        opts=None,
    ):
        return kind(
            cls._graph_class()(
                **filter_kwargs(opts, figure=figure, config=config, id=id)
            ),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def _datatable_class(cls):
        from dash_table import DataTable

        return DataTable

    @classmethod
    def date_picker_single_input(
        cls,
        date=None,
        label=Component.UNDEFINED,
        role="input",
        component_property="date",
        kind=Input,
        id=None,
        opts=None,
    ):
        if isinstance(date, datetime.date):
            date = date.isoformat()

        return kind(
            dcc.DatePickerSingle(date=date, **filter_kwargs(opts, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def date_picker_range_input(
        cls,
        start_date=None,
        end_date=None,
        label=Component.UNDEFINED,
        role="input",
        component_property=("start_date", "end_date"),
        kind=Input,
        id=None,
        opts=None,
    ):

        if isinstance(start_date, datetime.date):
            start_date = start_date.isoformat()
        if isinstance(end_date, datetime.date):
            end_date = end_date.isoformat()

        return kind(
            dcc.DatePickerRange(
                start_date=start_date, end_date=end_date, **filter_kwargs(opts, id=id)
            ),
            component_property=component_property,
            label=label,
            role=role,
        )



from .util import build_id
from ._callback import callback
from . import templates
import dash_labs.component_plugins
from . import plugins
from .dependency import Input, State, Output
from .version import __version__



from dash_labs import build_id, templates
from dash_labs.dependency import Output
from .base import ComponentPlugin


class DynamicLabelPlugin(ComponentPlugin):
    """
    Plugin that supports formatting a component's template label with the current
    component value
    """

    def __init__(self, input_dependency, template=None):
        if template is None:
            template = templates.FlatDiv()

        self.input_dependency = input_dependency
        self.format_string = self.input_dependency.label
        self.label_string = self.format_string.format(input_dependency.property_value())
        self.label_id = build_id("label")
        self.label_prop = template._label_value_prop

        self.input_dependency.label = self.label_string
        self.input_dependency.label_id = self.label_id

        # Init args and output
        args = dict(value=self.input_dependency)
        output = dict(
            label_value=Output(self.label_id, self.label_prop),
        )

        super(DynamicLabelPlugin, self).__init__(args=args, output=output)

    def get_value(self, args_value):
        """
        :param args_value: grouping of values corresponding to the dependency
            grouping returned by the args property
        :return: The value of the wrapped component
        """
        return args_value["value"]

    def get_output_values(self, args_value):
        return dict(label_value=self._format_label(self.get_value(args_value)))

    def _format_label(self, value):
        return self.format_string.format(value)



from functools import update_wrapper, wraps
from collections import OrderedDict

from dash import exceptions, Dash
from dash.development.base_component import Component
from .dependency import DashLabsDependency, State, Input, Output
from dash_labs.grouping import (
    flatten_grouping,
    make_grouping_by_index,
    validate_grouping,
    map_grouping,
)
from dash.dependencies import (
    Input as Input_dash,
    State as State_dash,
)
import dash_html_components as html

_ALL_DEPS = (Input, Output, State)


class CallbackWrapper:
    """
    Class that functions decorated by app.callback to provide access to
    original function and callback decorator arguments.
    """

    def __init__(
        self,
        fn,
        args_dependencies,
        output_dependencies,
        flat_fn=None,
        flat_output_deps=None,
        flat_input_deps=None,
        flat_state_deps=None,
    ):
        self.fn = fn
        self.args = args_dependencies
        self.output = output_dependencies
        self._flat_fn = flat_fn
        self._flat_output_deps = flat_output_deps
        self._flat_input_deps = flat_input_deps
        self._flat_state_deps = flat_state_deps
        update_wrapper(self, self.fn)

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)


def _normalize_inputs(inputs, state):
    # Handle positional inputs/state as int dict
    if inputs == [] and isinstance(state, dict):
        inputs = {}

    if not isinstance(inputs, dict):
        input_form = "list"
        if state is not None and isinstance(state, dict):
            raise ValueError("inputs and state must both be lists or dicts")

        if isinstance(inputs, tuple):
            inputs = list(inputs)
        elif not isinstance(inputs, list):
            inputs = [inputs]

        if state:
            if isinstance(state, tuple):
                state = list(state)
            elif not isinstance(state, list):
                state = [state]

        # Create dict from positional argument indices to dependency objects
        # Note that state values always come after inputs in Dash 1
        inputs = {i: val for i, val in enumerate(inputs)}
        num_inputs = len(inputs)
        if state is not None:
            state = {i + num_inputs: val for i, val in enumerate(state)}
    else:  # isinstance(inputs, dict):
        input_form = "dict"

    if state is None:
        state = {}

    # Check for duplicate keys
    dups = [k for k in inputs if k in state]
    if dups:
        raise ValueError(
            "argument names must be unique across input and state\n"
            "    The following were found in both: {dups}".format(dups=dups)
        )

    # Preprocess non-dependency inputs and state into arg instances
    all_inputs = OrderedDict()
    combined_inputs_state = inputs.copy()
    combined_inputs_state.update(state)
    for name, pattern in combined_inputs_state.items():
        if isinstance(pattern, DashLabsDependency) and not pattern.has_component:
            pattern = pattern.dependencies(labs=True)

        flat_deps = flatten_grouping(pattern)
        for dep in flat_deps:
            if not isinstance(dep, DashLabsDependency):
                raise ValueError("Invalid dependency: {}".format(dep))

            if dep.has_component:
                # Apply default label if undefined
                # (if user explicitly set label to None, leave it alone)
                if dep.label is Component.UNDEFINED:
                    dep.label = name

                # Update default role if undefined
                if dep.role is Component.UNDEFINED:
                    dep.role = "input"

        all_inputs[name] = pattern

    return all_inputs, input_form


def _normalize_output(output, template):
    # output_form stores whether wrapped function is expected to return values as
    # scalar, list, or dict.
    output_form = None
    if output is None or isinstance(output, list) and len(output) == 0:
        if template is None:
            raise ValueError(
                "No output dependency objects specified, and no template provided to\n"
                "construct a default output component."
            )
        else:
            output = template.default_output()

    if not isinstance(output, (list, dict)):
        output_form = "scalar"
        output = [output]

    if isinstance(output, list):
        # Convert output from list to
        output = {i: val for i, val in enumerate(output)}
        output_form = output_form or "list"
    else:
        output_form = output_form or "dict"

    all_output = OrderedDict()
    for name, pattern in output.items():
        if isinstance(pattern, DashLabsDependency) and not pattern.has_component:
            pattern = pattern.dependencies(labs=True)

        flat_deps = flatten_grouping(pattern)
        for dep in flat_deps:
            if not isinstance(dep, DashLabsDependency):
                raise ValueError("Invalid dependency: {}".format(dep))

            if dep.has_component:
                # Check is arg is holding a pattern
                if dep.label is Component.UNDEFINED:
                    dep.label = None

                if dep.role is Component.UNDEFINED:
                    dep.role = "output"

        all_output[name] = pattern

    return all_output, output_form


def _add_arg_components_to_template(vals, template):
    for name, val in vals.items():
        deps = flatten_grouping(val)
        for dep in deps:
            if dep.has_component:
                opts = {}
                if isinstance(name, str):
                    opts["name"] = name

                template.add_component(
                    component=dep.component_id,
                    component_property=dep.component_property,
                    role=dep.role,
                    label=dep.label,
                    label_id=dep.label_id,
                    **opts,
                )


def _get_arg_input_state_dependencies(all_inputs):
    input_groupings = OrderedDict()
    input_deps = []
    state_deps = []

    # Collect input groupings
    for name, val in all_inputs.items():
        if isinstance(val, Input):
            grouping = val.dependencies()
            flat_dependencies = val.flat_dependencies()
        else:
            flat_dependencies = [d.dependencies() for d in flatten_grouping(val)]
            if flat_dependencies and isinstance(flat_dependencies[0], Input_dash):
                grouping = val
            else:
                continue
        slc = slice(len(input_deps), len(input_deps) + len(flat_dependencies))
        input_groupings[name] = (grouping, slc)
        input_deps.extend(flat_dependencies)

    # Process state
    num_inputs = len(input_deps)
    for name, val in all_inputs.items():
        if isinstance(val, State):
            grouping = val.dependencies()
            flat_dependencies = val.flat_dependencies()
        else:
            flat_dependencies = [d.dependencies() for d in flatten_grouping(val)]
            if not flat_dependencies or isinstance(flat_dependencies[0], State_dash):
                grouping = val
            else:
                continue

        slc = slice(
            num_inputs + len(state_deps),
            num_inputs + len(state_deps) + len(flat_dependencies),
        )
        input_groupings[name] = (grouping, slc)
        state_deps.extend(flat_dependencies)

    return input_groupings, input_deps, state_deps


def _get_arg_output_dependencies(all_outputs):
    output_groupings = OrderedDict()
    output_deps = []

    # Collect input groupings
    for name, val in all_outputs.items():
        if isinstance(val, Output):
            grouping = val.dependencies()
            flat_dependencies = val.flat_dependencies()
        else:
            grouping = val
            flat_dependencies = [d.dependencies() for d in flatten_grouping(val)]

        output_groupings[name] = grouping
        output_deps.extend(flat_dependencies)

    return output_groupings, output_deps


def callback(app, *args, **kwargs):
    return _callback(
        app, *args, **kwargs, _wrapped_callback=getattr(Dash, "_wrapped_callback", None)
    )


def _callback(
    app,
    *_args,
    **_kwargs,
):
    """
    Implementation of the dash-labs app.callback
    """
    # Parse input arguments in a way that is consistent with Dash 1 app.callback
    # Note that the args argument to app.callback is merged with inputs
    (
        output,
        inputs,
        state,
        prevent_initial_callbacks,
        template,
        _wrapped_callback,
    ) = handle_callback_args(_args, _kwargs)

    # Combine inputs and stated and normalize into an OrderedDict. List/tuple inputs
    # have integer keys, which dict inputs have keyword argument names as keys.
    all_inputs, input_form = _normalize_inputs(inputs, state)

    # Likewise, normalize outputs into an OrderedDict
    all_outputs, output_form = _normalize_output(output, template)

    # For dependencies wrapping component instances, add component to template
    if template is not None:
        _add_arg_components_to_template(all_inputs, template)
        _add_arg_components_to_template(all_outputs, template)

    # Compute Dash 1 compatible Input/State dependency lists and argument groupings
    input_groupings, input_deps, state_deps = _get_arg_input_state_dependencies(
        all_inputs
    )

    # Compute Dash 1 compatible Output dependency list and output groupings
    output_groupings, output_deps = _get_arg_output_dependencies(all_outputs)

    def wrapped(fn) -> CallbackWrapper:
        # Preprocess arguments to callback function to create expected argument
        # groupings from the Dash 1 flat input arguments
        callback_fn = map_input_arguments(fn, input_groupings, input_form)

        # Postprocess return value of wrapped function to validate returned groupings
        # and flatten values into linear list expected by Dash 1 callback
        callback_fn = map_output_arguments(
            callback_fn,
            output_groupings,
            output_form,
        )

        # Register wrapped function with app.callback
        if _wrapped_callback is None:
            _callback = Dash.callback
        else:
            _callback = _wrapped_callback

        # Install callback
        if input_deps:
            _callback(
                app,
                output_deps,
                input_deps,
                state_deps,
                prevent_initial_call=prevent_initial_callbacks,
            )(callback_fn)

        # Return CallbackWrapper instance that mimicks wrapped function while providing
        # access to additional metadata about the callback.
        return CallbackWrapper(
            fn,
            all_inputs,
            all_outputs,
            callback_fn,
            output_deps,
            input_deps,
            state_deps,
        )

    return wrapped


def map_input_arguments(fn, input_groupings, input_form):
    @wraps(fn)
    def wrapper(*args):
        # args are the flat list of positional arguments passed to function by
        # app.callback

        expected_num_args = max([slc.stop for _, slc in input_groupings.values()])
        if len(args) != expected_num_args:
            # Not an error condition a user should reach since we control args and
            # input groupings
            raise ValueError(
                "Expected {} input value(s), received {}".format(
                    expected_num_args, len(args)
                )
            )

        fn_kwargs = {}
        for name, (grouping, slc) in input_groupings.items():
            fn_kwargs[name] = make_grouping_by_index(grouping, list(args[slc]))

        if input_form == "list":
            # keys of fn_kwargs are integers and we need to build list of position
            # arguments for fn
            fn_args = [None for _ in fn_kwargs]
            for i, val in fn_kwargs.items():
                fn_args[i] = val
            return fn(*fn_args)
        elif input_form == "scalar":
            return fn(fn_kwargs[0])
        else:  # "dict"
            return fn(**fn_kwargs)

    return wrapper


def map_output_arguments(fn, dep_groupings, output_form):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        res = fn(*args, **kwargs)
        output = []

        if output_form == "scalar":
            dep_grouping = dep_groupings[0]
            if isinstance(dep_grouping, tuple) and isinstance(res, list):
                # Accept list when expecting a tuple
                res = tuple(res)

            flat_res = extract_and_validate_output_values(res, dep_grouping)
            output.extend(flat_res)
        elif output_form == "list":
            if isinstance(res, tuple):
                # Accept tuple when expecting a list
                res = list(res)

            assert isinstance(res, list)
            assert len(res) == len(dep_groupings)

            for i in range(len(dep_groupings)):
                flat_res = extract_and_validate_output_values(res[i], dep_groupings[i])
                output.extend(flat_res)
        else:  # form == "dict"
            assert isinstance(res, dict)
            assert set(res) == set(dep_groupings)

            for name, grouping in dep_groupings.items():
                flat_res = extract_and_validate_output_values(
                    res[name], dep_groupings[name]
                )
                output.extend(flat_res)

        # The return value is the flat list of output arguments returned by the
        # function registered with app.callback
        return output

    return wrapper


def extract_and_validate_output_values(res_grouping, dep_grouping):
    # Extracting property values from dependency components
    if isinstance(res_grouping, DashLabsDependency):
        res_grouping = res_grouping.property_value()

    # Check value against schema
    validate_grouping(res_grouping, dep_grouping)
    flat_results = flatten_grouping(res_grouping, dep_grouping)

    return flat_results


def _validate_prop_name(component, prop_name):
    if (
        isinstance(prop_name, str)
        and prop_name.isidentifier()
        and prop_name in component._prop_names
    ):
        pass
    else:
        raise ValueError(
            "Invalid property {prop} received for component of type {typ}\n".format(
                prop=repr(prop_name), typ=type(component)
            )
        )


def _validate_prop_grouping(component, props):
    for prop in flatten_grouping(props):
        _validate_prop_name(component, prop)


# Modified Dash Callback Validation logic
# ---------------------------------------
def extract_callback_args(args, kwargs, names, type_):
    """Extract arguments for callback from a name and type"""
    parameters = [kwargs[name] for name in names if name in kwargs]

    if parameters:
        if len(parameters) > 1:
            raise ValueError("Only one of input and args may be specified")
        parameters = parameters[0]
        if not isinstance(parameters, (list, tuple, dict)):
            # accept a single item, not wrapped in a list, for any of the
            # categories as a named arg (even though previously only output
            # could be given unwrapped)
            return [parameters]
    else:
        while args and isinstance(args[0], (type_, type_.dependency_class)):
            arg = args.pop(0)
            parameters.append(arg)

    # Convert dash.dependency objects into dash-labs equivalents
    def to_dl(v):
        if isinstance(v, type_.dependency_class):
            return type_(v.component_id, v.component_property)
        else:
            return v

    if isinstance(parameters, list):
        parameters = [map_grouping(to_dl, p) for p in parameters]
    else:
        parameters = map_grouping(to_dl, parameters)

    return parameters


def handle_callback_args(args, kwargs):
    """Split args into outputs, inputs and states"""
    prevent_initial_call = kwargs.get("prevent_initial_call", None)
    if prevent_initial_call is None and args and isinstance(args[-1], bool):
        args, prevent_initial_call = args[:-1], args[-1]

    template = kwargs.get("template", None)
    _wrapped_callback = kwargs.get("_wrapped_callback", None)

    # flatten args, to support the older syntax where outputs, inputs, and states
    # each needed to be in their own list
    flat_args = []
    for arg in args:
        flat_args += arg if isinstance(arg, (list, tuple)) else [arg]

    outputs = extract_callback_args(flat_args, kwargs, ["output"], Output)
    validate_outputs = outputs
    if isinstance(outputs, (list, tuple)) and len(outputs) == 1:
        out0 = kwargs.get("output", args[0] if args else None)
        if not isinstance(out0, (list, tuple, dict)):
            # unless it was explicitly provided as a list, a single output
            # should be unwrapped. That ensures the return value of the
            # callback is also not expected to be wrapped in a list.
            outputs = outputs[0]

    inputs = extract_callback_args(flat_args, kwargs, ["inputs", "args"], Input)
    states = extract_callback_args(flat_args, kwargs, ["state"], State)

    input_values = inputs.values() if isinstance(inputs, dict) else inputs
    state_values = states.values() if isinstance(states, dict) else states
    output_values = (
        validate_outputs.values()
        if isinstance(validate_outputs, dict)
        else validate_outputs
    )

    validate_callback(output_values, input_values, state_values, flat_args)

    return outputs, inputs, states, prevent_initial_call, template, _wrapped_callback


def validate_callback(output, inputs, state, extra_args):
    is_multi = isinstance(output, (list, tuple))

    outputs = output if is_multi else [output]

    if extra_args:
        if not isinstance(extra_args[0], _ALL_DEPS):
            raise exceptions.IncorrectTypeException(
                """
                Positional callback arguments must be `Output`, `Input`, or `State` objects,
                optionally wrapped in a list or tuple. We found (possibly after
                unwrapping a list or tuple):
                {}
                """.format(
                    repr(extra_args[0])
                )
            )

        raise exceptions.IncorrectTypeException(
            """
            In a callback definition, you must provide all Outputs first,
            then all Inputs, then all States. After this item:
            {}
            we found this item next:
            {}
            """.format(
                repr((outputs + inputs + state)[-1]), repr(extra_args[0])
            )
        )



from dash_labs.grouping import flatten_grouping
from dash_labs.templates import FlatDiv


class ComponentPlugin:
    """
    Base class for component plugins
    """

    def __init__(self, args, output, template=None):
        """
        Superclass constructor that should be called by subclass constructors

        :param args: Grouping of dependency objects that specify the inputs required
            by the component plugin. If the plugin does not require inputs, this should
            be set to an empty tuple.
        :param output: Grouping of dependency objects that specify the outputs produced
            by the component plugin.
        :param template: Template for use by template. Defaults to FlatDiv.
        """
        self._args = args
        self._output = output

        if template is None:
            template = FlatDiv()
        self.template = template

    def get_output_values(self, args_value):
        """
        Method that produces a grouping of the plugin's output values, which correspond
        to the dependency grouping returned by the args property

        :param args_value: grouping of values corresponding to the dependency
            grouping returned by the args property.
        :return: grouping of values corresponding to the dependency grouping returned
            by the output property
        """
        raise NotImplementedError

    @property
    def args(self):
        """
        :return: Grouping of Input/State dependency objects for the inputs required by
            the plugin's get_output_values method
        """
        return self._args

    @property
    def output(self):
        """
        :return: Grouping of Output dependency objects for the callback output values
            produced by the plugin's get_output_values method
        """
        return self._output

    def install_callback(self, app):
        """
        Convenience method to install the minimal required callback definition to
        use the plugin. This is useful for cases where the plugin doesn't need to
        integrate with any other components.

        :param app: dash.Dash app instance
        """

        @app.callback(
            args=[self.args],
            output=[self.output],
            template=self.template,
        )
        def callback(plugin_inputs):
            return [self.get_output_values(plugin_inputs)]

    @property
    def args_components(self):
        """
        :return: list of the components corresponding to the plugin's args dependencies
        """
        return [
            self.template.build_argument_components(
                c.component_id, label=c.label, label_id=c.label_id
            ).container_component
            for c in flatten_grouping(self.args)
            if c.has_component
        ]

    @property
    def output_components(self):
        """
        :return: list of the components corresponding to the plugin's output
            dependencies
        """
        return [
            self.template.build_argument_components(
                c.component_id, label=c.label, label_id=c.label_id
            ).container_component
            for c in flatten_grouping(self.output)
            if c.has_component
        ]

    @property
    def container(self):
        """
        :return: Collection of all arg and output components, stored in the default
            component container for the current template (typically an html.Div)
        """
        return self.template.default_output(
            children=self.args_components + self.output_components
        ).component_id



import numpy as np

try:
    import spectra
    from colormath.density import auto_density
    from colormath.color_objects import LabColor as LabColor
    from colormath.color_diff import delta_e_cie1994
except ImportError:
    msg = (
        "Generating plotly.py figure templates from bootstrap theme files requires\n"
        "the optional spectra package, which can be installed using pip...\n"
        "    $ pip install spectra\n"
        "or conda...\n"
        "    $ conda install -c conda-forge spectra"
    )
    raise ValueError(msg)


white = spectra.lab(100, 0, 0)
black = spectra.lab(0, 0, 0)


def to_colormath(spectra_color):
    lab_values = spectra.html(spectra_color.hexcode).to("lab").values
    return LabColor(*lab_values)


def color_distance(clr1, clr2):
    return delta_e_cie1994(to_colormath(clr1), to_colormath(clr2))


# Distance matric
def color_distance_matrix(colors):
    return np.array(
        [
            [color_distance(c1, c2) for i, c1 in enumerate(colors)]
            for j, c2 in enumerate(colors)
        ],
        dtype="float32",
    )


def get_darkened_colors(colors, darkening_list):
    return [c.darken(d) for c, d in zip(colors, darkening_list)]


def best_darkening(c1, c2, c1_step=(1, 1), c2_step=(1, 1)):
    d = color_distance(c1, c2)
    d1 = color_distance(c1.darken(c1_step[1]), c2)
    dm1 = color_distance(c1.darken(c1_step[0]), c2)
    d2 = color_distance(c1, c2.darken(c2_step[1]))
    dm2 = color_distance(c1, c2.darken(c2_step[0]))

    # Return step with sign that increases distance most
    # Return 0 if either step lower distance
    return sorted(
        [
            (d, (0, 0)),
            (d1, (c1_step[1], 0)),
            (dm1, (c1_step[0], 0)),
            (d2, (0, c2_step[1])),
            (dm2, (0, c2_step[0])),
        ]
    )[-1]


def separate_colorway(html_colors):

    try:
        raw_colors = [
            spectra.rgb(*[c / 255 for c in to_rgb_tuple(clr)]) for clr in html_colors
        ]
    except ValueError:
        # Unable to parse colors as hex or rgb, return as-is
        return html_colors

    test_colors = [white] + raw_colors + [black]

    darkenings = list(np.zeros(len(test_colors)))
    threshold = 36

    max_shift = 16
    max_step = 16
    max_iterations = 4
    max_step_factor = 0.9

    iterations = 0
    distances = np.ones((len(html_colors) + 2, len(html_colors) + 2)) * np.nan

    while iterations < max_iterations:
        for i in range(len(test_colors) - 1):
            c1 = test_colors[i].darken(darkenings[i])
            for j in range(i + 1, len(test_colors)):
                c2 = test_colors[j].darken(darkenings[j])
                distance = color_distance(c1, c2)
                distances[i, j] = distance

                # When comparing to black and white,
                # skip if at least threshold units away
                if distance > threshold:
                    continue

                # Compute max step based on how close colors are
                this_step = max_step * ((100 - distance) / 100) ** 2

                # Clamp max steps based on how close we are to max shift allowances
                c1_step_up = max(0, min(this_step, max_shift - darkenings[i]))
                c2_step_up = max(0, min(this_step, max_shift - darkenings[j]))
                c1_step_down = min(0, max(-this_step, -darkenings[i] - max_shift))
                c2_step_down = min(0, max(-this_step, -darkenings[j] - max_shift))

                # Compute best way to lighten or darken ONE of the colors (not both)
                distance, (delta1, delta2) = best_darkening(
                    c1,
                    c2,
                    c1_step=(c1_step_down, c1_step_up),
                    c2_step=(c2_step_down, c2_step_up),
                )
                distances[i, j] = distance

                darkenings[i] += delta1
                darkenings[j] += delta2

        iterations += 1
        max_step *= max_step_factor

    result = [clr.hexcode for clr in get_darkened_colors(test_colors, darkenings)[1:-1]]

    return result


def hex_to_rgb(clr):
    clr = clr.lstrip("#")
    if len(clr) == 3:
        clr = "".join(c[0] * 2 for c in clr)
    return tuple(int(clr[i : i + 2], 16) for i in (0, 2, 4))


def to_rgb_tuple(color):
    from plotly.colors import unlabel_rgb

    if isinstance(color, tuple):
        pass
    elif color.startswith("#"):
        color = hex_to_rgb(color)
    else:
        color = unlabel_rgb(color)

    return tuple(int(c) for c in color)


def make_grid_color(bg_color, font_color, weight=0.1):
    bg_color = to_rgb_tuple(bg_color)
    font_color = to_rgb_tuple(font_color)

    s_bg_color = spectra.rgb(*[c / 255 for c in bg_color])
    s_font_color = spectra.rgb(*[c / 255 for c in font_color])
    return s_bg_color.blend(s_font_color, weight).hexcode


def maybe_blend(base_color, overlay_color):
    """
    Try to blend semi transparent overlay color on opaque
    base color. Return None if not successful
    """
    import re

    try:
        bc = spectra.html(base_color).to("rgb")
    except ValueError:
        return None

    try:
        # If overlay color is hex code or named color, it's
        # opaque, return as is
        return spectra.html(overlay_color).hexcode
    except ValueError:
        # Otherwise, it might be rgba
        pass

    rgba_match = re.match(r"rgba\(([^,]+),([^,]+),([^,]+),([^,]+)\)", overlay_color)
    if rgba_match is None:
        return None

    r, g, b, a = [float(n) for n in rgba_match.groups()]
    overlay_rgb = spectra.rgb(r / 255, g / 255, b / 255)
    blended = overlay_rgb.blend(bc, 1 - a)
    return blended.hexcode



from dash_labs import build_id
from dash_labs.dependency import Output, Input
from dash_labs.templates import FlatDiv
from .base import ComponentPlugin
import plotly.graph_objects as go
import dash


class GreyscaleImageROI(ComponentPlugin):
    """
    Plugin that supports drawing, and editing, a rectangular box on a greyscale image
    and retrieving the selected region and selected pixels.
    """

    def __init__(
        self,
        img,
        template=None,
        image_label=None,
        newshape=None,
        title=None,
        fig_update_callback=None,
    ):
        import plotly.express as px

        if template is None:
            template = FlatDiv()

        self.img = img

        if newshape is None:
            newshape = dict(
                fillcolor="lightgray", opacity=0.2, line=dict(color="black", width=8)
            )

        self.newshape = newshape
        self.fig_update_callback = fig_update_callback

        top_margin = 60 if title is not None else 20

        self.image_fig = px.imshow(img, binary_string=True).update_layout(
            dragmode="drawrect",
            margin=dict(l=20, b=20, r=20, t=top_margin),
            newshape=self.newshape,
            title_text=title,
        )

        self.image_graph_id = build_id("image-graph")
        self.histogram_graph_id = build_id("histogram-graph")
        self.image_label = image_label

        # Initialize args component dependencies
        args = template.graph_output(
            self.image_fig,
            component_property="relayoutData",
            kind=Input,
            role="input",
            id=self.image_graph_id,
            label=self.image_label,
        )

        output = {"image_figure": Output(self.image_graph_id, "figure")}

        super().__init__(args, output, template)

    def get_output_values(self, args_value, title=None):
        relayout_data = args_value
        if relayout_data:
            # shape coordinates are floats, we need to convert to ints for slicing
            bounds = self._extract_pixel_bounds_from_shape(relayout_data)
            if bounds is None:
                return {
                    "image_figure": dash.no_update,
                }

            x0, y0, x1, y1 = bounds
            shapes = self._make_rect(x0, y0, x1, y1)

            top_margin = 60 if title is not None else 20

            new_image_fig = go.Figure(self.image_fig)
            new_image_fig.update_layout(
                shapes=shapes,
                margin_t=top_margin,
            )

            if title is not None:
                new_image_fig.update_layout(title_text=title)

            return {
                "image_figure": new_image_fig,
            }
        else:
            return {
                "image_figure": dash.no_update,
            }

    def get_rect_bounds(self, args_value, integer=True):
        """
        :param args_value: grouping of values corresponding to the dependency
            grouping returned by the args property
        :param integer: If True (default), convert bounds into integer pixel
            coordinates. If False, return floating point bounds
        :return: Bounds tuple of the form (x0, y0, x1, y1), or None if no box shape
            is present
        """
        relayout_data = args_value
        if relayout_data:
            bounds = self._extract_pixel_bounds_from_shape(relayout_data)
            if integer and bounds is not None:
                bounds = tuple([int(b) if b is not None else None for b in bounds])
            return bounds
        else:
            return None

    def get_image_slice(self, args_value):
        """
        :param args_value: grouping of values corresponding to the dependency
            grouping returned by the args property
        :return: Slice of the original image that is inside the current bounds, or
            None if no box shape present
        """
        relayout_data = args_value
        if relayout_data:
            bounds = self._extract_pixel_bounds_from_shape(relayout_data)

            if bounds is not None:
                x0, y0, x1, y1 = bounds
                return self.img[int(y0) : int(y1), int(x0) : int(x1)]
            else:
                return None
        else:
            return None

    def _make_rect(self, x0, y0, x1, y1):
        if all((x0, y0, x1, y1)):
            return [
                dict(
                    {
                        "editable": True,
                        "type": "rect",
                        "x0": x0,
                        "y0": y0,
                        "x1": x1,
                        "y1": y1,
                    },
                    **self.newshape
                )
            ]
        else:
            return []

    def _extract_pixel_bounds_from_shape(self, relayout_data):
        x0, y0, x1, y1 = (None,) * 4
        if "shapes" in relayout_data:
            shape = relayout_data["shapes"][-1]
            x0, y0 = shape["x0"], shape["y0"]
            x1, y1 = shape["x1"], shape["y1"]
        elif any(["shapes" in key for key in relayout_data]):
            x0 = [relayout_data[key] for key in relayout_data if "x0" in key][0]
            x1 = [relayout_data[key] for key in relayout_data if "x1" in key][0]
            y0 = [relayout_data[key] for key in relayout_data if "y0" in key][0]
            y1 = [relayout_data[key] for key in relayout_data if "y1" in key][0]

        # normalize coordinates and clamp to valid image boundaries
        if x0 and x1:
            if x0 > x1:
                x0, x1 = x1, x0
            x0 = 0 if x0 < 0 else x0
            x1 = self.img.shape[1] if x1 > self.img.shape[1] else x1
        if y0 and y1:
            if y0 > y1:
                y0, y1 = y1, y0
            y0 = 0 if y0 < 0 else y0
            y1 = self.img.shape[0] if y1 > self.img.shape[0] else y1

        if all(b is None for b in (x0, y0, x1, y1)):
            return None
        else:
            return x0, y0, x1, y1



from dash_labs.templates.base import BaseTemplate
import dash_html_components as html


class FlatDiv(BaseTemplate):
    """
    Trivial template that returns all input containers in a single flat Div,
    inputs followed by outputs
    """

    def __init__(self):
        super().__init__()

    def _perform_layout(self):
        children = []
        children.extend(self.get_containers("input"))
        children.extend(self.get_containers("output"))
        return html.Div(children=children)



from collections import OrderedDict

from dash.development.base_component import Component

from dash_labs.templates.base import BaseTemplate
import dash_html_components as html
from dash_labs.util import filter_kwargs, build_id
from dash_labs.dependency import Input
import plotly.graph_objects as go
import copy
import plotly.io as pio


class BaseDbcTemplate(BaseTemplate):
    """
    Base class for templates based on the dash-bootstrap-components library
    """

    # - Undo the negative margins that table rows pick up from bootstrap's own row
    #   CSS class. Otherwise, table expand in width outside of their container.
    # - Style various elements using bootstrap css variables
    _inline_css = (
        BaseTemplate._inline_css
        + """

         .dash-spreadsheet .row {
            margin-left: 0;
            margin-right: 0;
         }
         
         .dash-spreadsheet-inner input {
            color: var(--white) !important;
            font-weight: 800 !important;
         }         

         .dash-spreadsheet .dash-header {
            background-color: var(--primary) !important;
            color: var(--white) !important;
            border-color: rgba(128, 128, 128, 0.3) !important;
            font-weight: 400 !important;
            font-size: 1.25em !important;
         }
         
         .dash-spreadsheet .dash-filter {
            background-color: var(--primary) !important;
            border-color: rgba(128, 128, 128, 0.3) !important;
         }
         
         .dash-spreadsheet .dash-filter input {
            color: var(--white) !important;
         }
         
         .dash-spreadsheet td {
            background-color: var(--light) !important;
            border-color: rgba(128, 128, 128, 0.3) !important;
            color: var(--dark) !important;
         }
         
         .dash-table-container .dash-spreadsheet-container .dash-spreadsheet-inner table {
            --hover: transparent !important;
         }
         
         .dash-table-container .dash-spreadsheet-container .dash-spreadsheet-inner td.focused {
            background-color: var(--info) !important;
            color: white !important;
         }
                 
         .dash-spreadsheet td.cell--selected {
            background-color: var(--info) !important;
            color: white !important;
         }

         .tab-pane .card {
            border-top-left-radius: 0;
         }
         
         .nav-tabs {
            border-bottom-width: 0 !important;
         }
         
         .nav-link {
            border-width: 0.5px !important;
            border-color: rgba(100, 100, 100, 0.4) !important;
            color: var(--primary) !important;
         }
         
         .nav-link.active {
            background-color: var(--primary) !important;
            color: var(--white) !important;
            border-color: rgba(100, 100, 100, 0.4);
         }
         
         .nav-item {
            margin-right: 0 !important;
         }
         
         .card {
            margin-bottom: 1rem !important;
            border-color: rgba(100, 100, 100, 0.4);
        }
        
    
        .rc-slider-handle {
            border: solid 2px var(--primary) !important;
        }
        
        .rc-slider-handle:focus {
            box-shadow: unset; !important;
            background-color: var(--primary);
        }
        
        .rc-slider-track {
            background-color: var(--primary);
            height: 2px;
            margin-top: 1px;
        }
        .rc-slider-rail {
            height: 2px;
            margin-top: 1px;
        }
        """
    )

    def __init__(self, theme=None, figure_template=False):
        """
        :param theme: Path to a bootstrap theme css file. If not provided, a default
            bootstrap theme will be used.
        :param figure_template: If True, generate a plotly.py figure template from the
            provided (or default) bootstrap theme. Figure templates will adopt the
            colors and fonts of the associated bootstrap theme css file.
        """
        super().__init__()
        self.theme = theme
        self.figure_template = figure_template

    # Methods designed to be overridden by subclasses
    @classmethod
    def build_labeled_component(cls, component, label, label_id=None, role=None):
        import dash_bootstrap_components as dbc

        if not label_id:
            label_id = build_id("label")

        label_component = dbc.Label(
            id=label_id,
            children=[label],
            style={"display": "block"},
            className="h5",
        )
        container_id = build_id("container")
        container = dbc.FormGroup(
            id=container_id, children=[label_component, component]
        )
        return container, "children", label_component, "children"

    @classmethod
    def build_containered_component(cls, component, role=None):
        import dash_bootstrap_components as dbc

        container_id = build_id("container")
        container = dbc.FormGroup(id=container_id, children=[component])
        return container, "children"

    def _configure_app(self, app):
        super()._configure_app(app)
        import dash_bootstrap_components as dbc

        # Check if there are any bootstrap css themes already added
        add_theme = True
        theme = self.theme
        for url in app.config.external_stylesheets:
            if "bootstrapcdn" in url:
                add_theme = False
                if theme is None:
                    theme = url
                break

        if add_theme:
            if self.theme is None:
                theme = dbc.themes.FLATLY
            else:
                theme = self.theme

            app.config.external_stylesheets.append(theme)

        if self.figure_template:
            self.make_figure_theme(theme, activate=True)

    @classmethod
    def make_figure_theme(cls, theme, activate=True, raise_on_failure=True):
        template = _try_build_plotly_template_from_bootstrap_css_path(theme)
        if template is None and raise_on_failure:
            raise ValueError(
                f"Failed to construct plotly.py figure theme from bootstrap theme:\n"
                f"   {theme}"
            )
        if activate:
            pio.templates["dash_bootstrap"] = template
            pio.templates.default = "dash_bootstrap"

        return template

    # component dependency constructors
    @classmethod
    def button_input(
        cls,
        children,
        color="secondary",
        size="md",
        outline=False,
        label=Component.UNDEFINED,
        role=Component.UNDEFINED,
        component_property="n_clicks",
        kind=Input,
        id=None,
        opts=None,
    ):
        import dash_bootstrap_components as dbc

        return kind(
            dbc.Button(
                children=children,
                **filter_kwargs(opts, id=id, color=color, size=size, outline=outline),
            ),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def dropdown_input(
        cls,
        options,
        value=Component.UNDEFINED,
        clearable=False,
        label=Component.UNDEFINED,
        role=Component.UNDEFINED,
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        import dash_bootstrap_components as dbc

        if isinstance(options, list) and options and not isinstance(options[0], dict):
            options = [{"value": opt, "label": opt} for opt in options]

        if clearable:
            options.insert(0, {"value": "", "label": "<Clear>"})

        # Set starting value if not specified
        if value is Component.UNDEFINED and options:
            value = options[0]["value"]

        # classname
        opts = opts or {}
        opts.setdefault("className", "h6")

        return kind(
            dbc.Select(options=options, **filter_kwargs(opts, value=value, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def textbox_input(
        cls,
        value=None,
        label=Component.UNDEFINED,
        role=Component.UNDEFINED,
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        import dash_bootstrap_components as dbc

        # classname
        opts = opts or {}
        opts.setdefault("className", "h6")

        return kind(
            dbc.Input(value=value, **filter_kwargs(opts, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )

    @classmethod
    def checklist_input(
        cls,
        options,
        value=None,
        label=Component.UNDEFINED,
        role=Component.UNDEFINED,
        component_property="value",
        kind=Input,
        id=None,
        opts=None,
    ):
        import dash_bootstrap_components as dbc

        if isinstance(options, list) and options and not isinstance(options[0], dict):
            options = [{"value": opt, "label": opt} for opt in options]

        opts = opts or {}

        return kind(
            dbc.Checklist(options=options, **filter_kwargs(opts, value=value, id=id)),
            component_property=component_property,
            label=label,
            role=role,
        )


class DbcCard(BaseDbcTemplate):
    def __init__(
        self,
        title=None,
        columns=6,
        min_width=400,
        height=None,
        margin="10px 10px 10px 10px",
        **kwargs,
    ):
        """
        Template that places all components in a single card

        :param title: Card title
        :param columns: Responsive width of card in columns (out of 12 columns)
        :param min_width: Minimum card width in pixels
        :param height: Fixed height of card or None (default) to let card's height
            expand to contents
        :param margin: CSS margin around row
        """
        super().__init__(**kwargs)
        self.title = title
        self.columns = columns
        self.height = height
        self.min_width = min_width
        self.margin = margin

    def _perform_layout(self):
        import dash_bootstrap_components as dbc

        # No callbacks here. Must be constant or idempotent
        card_children = []
        if self.title:
            card_children.append(dbc.CardHeader(self.title, className="h4"))

        card_body_children = []
        output_containers = self.get_containers("output")
        input_containers = self.get_containers("input")
        card_body_children.extend(output_containers)
        if output_containers and input_containers:
            card_body_children.append(html.Hr())
        card_body_children.extend(input_containers)

        card_children.append(dbc.CardBody(children=card_body_children))

        card_style = {"padding": 0}
        if self.margin:
            card_style["margin"] = self.margin
        if self.height is not None:
            card_style["height"] = self.height

        if self.min_width is not None:
            card_style["min-width"] = self.min_width

        class_name_kwarg = {}
        if self.columns is not None:
            class_name_kwarg["className"] = f"col-{int(self.columns)}"

        return dbc.Card(
            style=card_style,
            children=card_children,
            **class_name_kwarg,
        )


class DbcRow(BaseDbcTemplate):
    def __init__(
        self,
        title=None,
        row_height=None,
        input_cols=4,
        min_input_width="300px",
        margin="10px 0 10px 0",
        **kwargs,
    ):
        """
        Template that places inputs and outputs in separate cards, arranged in a row

        :param title: Input card title
        :param input_cols: Responsive width of input card in columns (out of 12 columns)
        :param min_input_width: Minimum input card width in pixels
        :param margin: CSS margin around row
        :param row_height: Fixed height of cards in the row in pixels.
            If None (defualt) each card will independently determine height based on
            contents.
        """
        super().__init__(**kwargs)
        self.title = title
        self.row_height = row_height
        self.input_cols = input_cols
        self.min_input_width = min_input_width
        self.margin = margin

    def _perform_layout(self):
        import dash_bootstrap_components as dbc

        output_card_children = []
        if self.title:
            output_card_children.append(dbc.CardHeader(self.title, className="h4"))

        output_card_children.append(dbc.CardBody(self.get_containers("output")))

        input_card_style = {}
        if self.min_input_width is not None:
            input_card_style["min-width"] = self.min_input_width

        row_style = {}
        if self.margin:
            row_style["margin"] = self.margin
        if self.row_height is not None:
            row_style["height"] = self.row_height

        class_name_kwarg = {}
        if self.input_cols is not None:
            class_name_kwarg["className"] = f"col-{int(self.input_cols)}"

        return dbc.Row(
            style=row_style,
            children=[
                dbc.Col(
                    children=dbc.Card(children=self.get_containers("input"), body=True),
                    style=input_card_style,
                    **class_name_kwarg,
                ),
                dbc.Col(dbc.Card(children=output_card_children)),
            ],
        )


class DbcSidebar(BaseDbcTemplate):
    def __init__(self, title=None, sidebar_columns=4, **kwargs):
        """
        Template that includes a title bar, places inputs in a sidebar, and outputs in
        a responsive card

        :param title: Title bar title string
        :param sidebar_columns: Responsive width of input card in columns
            (out of 12 columns)
        """
        super().__init__(**kwargs)
        self.title = title
        self.sidebar_columns = sidebar_columns

    def _perform_layout(self):
        import dash_bootstrap_components as dbc

        children = []

        if self.title:
            children.extend(
                [
                    html.H3(
                        className="bg-primary text-white",
                        children=self.title,
                        style={"padding": "0.5rem"},
                    ),
                ]
            )

        sidebar_card_style = {"border-radius": 0}

        row = dbc.Row(
            align="top",
            style={"padding": "10px", "margin": "0"},
            children=[
                dbc.Col(
                    children=dbc.Card(
                        children=self.get_containers("input"),
                        body=True,
                    ),
                    style=sidebar_card_style,
                    md=self.sidebar_columns,
                ),
                dbc.Col(
                    children=dbc.Card(
                        children=self.get_containers("output"), body=True
                    ),
                    md=12 - self.sidebar_columns,
                ),
            ],
        )
        children.append(row)
        return children

    @classmethod
    def _wrap_full_layout(cls, layout):
        import dash_bootstrap_components as dbc

        return dbc.Container(layout, fluid=True, style={"padding": 0})


class DbcSidebarTabs(BaseDbcTemplate):
    def __init__(self, tab_roles, title=None, sidebar_columns=4, **kwargs):
        """
        Template that includes a title bar, places inputs in a sidebar, and outputs in
        a set of tabs.

        :param tab_roles: List or dict of strings where each string specifies the name
            of the role corresponding to a single tab. If a list, the role name is
            also be used as the title of the corresponding tab. If a dict, the keys
            become the roles and the values become the tab labels
        :param title: Title bar title string
        :param sidebar_columns: Responsive width of input card in columns
            (out of 12 columns)
        """
        import dash_bootstrap_components as dbc

        # Set valid roles before constructor
        self.title = title
        self.sidebar_columns = sidebar_columns
        if isinstance(tab_roles, (list, tuple)):
            self.tab_roles = OrderedDict([(role, role) for role in tab_roles])
        else:
            self.tab_roles = OrderedDict(tab_roles)

        self._valid_roles = ["input", "output"] + list(self.tab_roles.keys())

        first_tab = next(iter(self.tab_roles))
        self._tabs = dbc.Tabs(id=build_id("tabs"), active_tab=first_tab)

        super().__init__(**kwargs)

    def _perform_layout(self):
        import dash_bootstrap_components as dbc

        children = []

        if self.title:
            children.extend(
                [
                    html.H3(
                        className="bg-primary text-white",
                        children=self.title,
                        style={"padding": "0.5rem"},
                    ),
                ]
            )

        sidebar_card_style = {"border-radius": 0}

        self._tabs.children = [
            dbc.Tab(
                [
                    dbc.Card(
                        children=list(reversed(self.get_containers(role))), body=True
                    ),
                ],
                tab_id=role,
                label=title,
            )
            for role, title in self.tab_roles.items()
        ]

        row = dbc.Row(
            align="top",
            style={"padding": "10px", "margin": "0"},
            children=[
                dbc.Col(
                    children=dbc.Card(
                        children=self.get_containers("input"),
                        body=True,
                    ),
                    style=sidebar_card_style,
                    md=self.sidebar_columns,
                ),
                dbc.Col(self._tabs, md=12 - self.sidebar_columns),
            ],
        )
        children.append(row)
        return children

    @classmethod
    def _wrap_full_layout(cls, layout):
        import dash_bootstrap_components as dbc

        return dbc.Container(layout, fluid=True, style={"padding": 0})

    def tab_input(self, kind=Input):
        """
        Dependency object that can be used to input the active tab
        :param kind: The dependency kind to return. One of dl.Input (default) or
            dl.State.
        :return: Dependency object referencing the active tab
        """
        return kind(self._tabs.id, "active_tab")


def _parse_rules_from_bootstrap_css(css_text):
    import tinycss2

    tinycss_parsed = tinycss2.parse_stylesheet(css_text)

    # Build dict from css selectors to dict of css prop-values
    rule_props = {}
    for token in tinycss_parsed:
        if token.type != "qualified-rule":
            continue
        rule = token
        selector_str = "".join([t.serialize() for t in rule.prelude])
        selectors = tuple(s.strip() for s in selector_str.split(","))
        property_strings = [
            entry
            for entry in "".join([c.serialize().strip() for c in rule.content]).split(
                ";"
            )
            if entry
        ]

        property_pairs = [prop_str.split(":") for prop_str in property_strings]
        for selector in selectors:
            for prop_pair in property_pairs:
                if len(prop_pair) != 2:
                    continue
                rule_props.setdefault(selector, {})
                prop_key = prop_pair[0]
                prop_value = prop_pair[1].replace("!important", "").strip()
                rule_props[selector][prop_key] = prop_value

    return rule_props


# Get title font color
def _get_font(rule_props):
    color = "#000"
    family = "sans-serif"

    for el in ["html", "body", "h1"]:
        color = rule_props.get(el, {}).get("color", color)
        family = rule_props.get(el, {}).get("font-family", family)

    return color, family


def _get_role_colors(rule_props):
    # Initialize role_colors with default values
    role_colors = {
        "primary": "#007bff",
        "secondary": "#6c757d",
        "success": "#28a745",
        "info": "#17a2b8",
        "warning": "#ffc107",
        "danger": "#dc3545",
        "light": "#f8f9fa",
        "dark": "#343a40",
    }

    # Override with role colors for current theme
    for prop, val in rule_props[":root"].items():
        if prop.startswith("--"):
            maybe_color = prop.lstrip("-")
            if maybe_color in role_colors:
                role_colors[maybe_color] = val

    return role_colors


def _build_plotly_template_from_bootstrap_css_text(css_text):
    from dash_labs.templates._colors import (
        make_grid_color,
        separate_colorway,
        maybe_blend,
    )

    # Parse css text
    rule_props = _parse_rules_from_bootstrap_css(css_text)

    # Initialize role_colors with default values
    role_colors = _get_role_colors(rule_props)

    # Get font info
    font_color, font_family = _get_font(rule_props)

    # Get background color
    plot_bgcolor = rule_props["body"].get("background-color", "#fff")
    paper_bgcolor = rule_props[".card"].get("background-color", plot_bgcolor)

    blended = maybe_blend(plot_bgcolor, paper_bgcolor)
    if blended is None:
        # Can't blend, use background color for everything
        paper_bgcolor = plot_bgcolor
    else:
        paper_bgcolor = blended

    # Build colorway
    colorway_roles = [
        "primary",
        "danger",
        "success",
        "warning",
        "info",
    ]
    colorway = [role_colors[r] for r in colorway_roles]
    colorway = separate_colorway(colorway)
    print("colorway", colorway)

    # Build grid color
    gridcolor = make_grid_color(plot_bgcolor, font_color, 0.08)

    # Make template
    template = copy.deepcopy(pio.templates["plotly_dark"])

    layout = template.layout
    layout.colorway = colorway
    layout.piecolorway = colorway
    layout.paper_bgcolor = paper_bgcolor
    layout.plot_bgcolor = plot_bgcolor
    layout.font.color = font_color
    layout.font.family = font_family
    layout.xaxis.gridcolor = gridcolor
    layout.yaxis.gridcolor = gridcolor
    layout.xaxis.gridwidth = 0.5
    layout.yaxis.gridwidth = 0.5
    layout.xaxis.zerolinecolor = gridcolor
    layout.yaxis.zerolinecolor = gridcolor
    layout.margin = dict(l=0, r=0, b=0)

    template.data.scatter = (go.Scatter(marker_line_color=plot_bgcolor),)
    template.data.scattergl = (go.Scattergl(marker_line_color=plot_bgcolor),)

    return template


def _try_build_plotly_template_from_bootstrap_css_path(css_url):
    import requests
    from urllib.parse import urlparse

    parse_result = urlparse(css_url)
    if parse_result.scheme:
        # URL
        response = requests.get(css_url)
        if response.status_code != 200:
            return None
        css_text = response.content.decode("utf8")
    elif parse_result.path:
        # Local file
        with open(parse_result.path, "rt") as f:
            css_text = f.read()

    return _build_plotly_template_from_bootstrap_css_text(css_text)



"""
This module contains a collection of utility function for dealing with property
groupings.

Terminology:

For the purpose of grouping and ungrouping, tuples and dictionaries are considered
"composite values" and all other values are considered "scalar values".

A "grouping value" is either composite or scalar.

A "schema" is a grouping value that can be used to encode an expected grouping
structure

"""


def flatten_grouping(grouping, schema=None):
    """
    Convert a grouping value to a list of scalar values

    :param grouping: grouping value to flatten
    :param schema: If provided, a grouping value representing the expected structure of
        the input grouping value. If not provided, the grouping value is its own schema.
        A schema is required in order to be able treat tuples and dicts in the input
        grouping as scalar values.

    :return: list of the scalar values in the input grouping
    """
    if schema is None:
        schema = grouping

    if isinstance(schema, tuple):
        return [
            g
            for group_el, schema_el in zip(grouping, schema)
            for g in flatten_grouping(group_el, schema_el)
        ]
    elif isinstance(schema, dict):
        return [
            g
            for group_el, schema_el in zip(grouping.values(), schema.values())
            for g in flatten_grouping(group_el, schema_el)
        ]
    else:
        return [grouping]


def grouping_len(grouping):
    """
    Get the length of a grouping. The length equal to the number of scalar values
    contained in the grouping, which is equivalent to the length of the list that would
    result from calling flatten_grouping on the grouping value.

    :param grouping: The grouping value to calculate the length of
    :return: non-negative integer
    """
    if isinstance(grouping, tuple):
        return sum([grouping_len(group_el) for group_el in grouping])
    elif isinstance(grouping, dict):
        return sum([grouping_len(group_el) for group_el in grouping.values()])
    else:
        return 1


def make_grouping_by_index(schema, flat_values):
    """
    Make a grouping like the provided grouping schema, with scalar values drawn from a
    flat list by index.

    Note: Scalar values in schema are not used

    :param schema: Grouping value encoding the structure of the grouping to return
    :param flat_values: List of values with length matching the grouping_len of schema.
        Elements of flat_values will become the scalar values in the resulting grouping
    """

    def _perform_make_grouping_like(value, next_values):
        if isinstance(value, tuple):
            return tuple(
                _perform_make_grouping_like(el, next_values)
                for i, el in enumerate(value)
            )
        elif isinstance(value, dict):
            return {
                k: _perform_make_grouping_like(v, next_values)
                for i, (k, v) in enumerate(value.items())
            }
        else:
            return next_values.pop(0)

    if not isinstance(flat_values, list):
        raise ValueError(
            "The flat_values argument must be a list. "
            "Received value of type {typ}".format(typ=type(flat_values))
        )

    expected_length = len(flatten_grouping(schema))
    if len(flat_values) != expected_length:
        raise ValueError(
            "The specified grouping pattern requires {n} elements but received {m}\n"
            "    Grouping patter: {pattern}\n"
            "    Values: {flat_values}".format(
                n=expected_length,
                m=len(flat_values),
                pattern=repr(schema),
                flat_values=flat_values,
            )
        )

    return _perform_make_grouping_like(schema, list(flat_values))


def map_grouping(fn, grouping):
    """
    Map a function over all of the scalar values of a grouping, maintaining the
    grouping structure

    :param fn: Single-argument function that accepts and returns scalar grouping values
    :param grouping: The grouping to map the function over
    :return: A new grouping with the same structure as input grouping with scalar
        values updated by the input function.
    """
    if isinstance(grouping, tuple):
        return tuple(map_grouping(fn, g) for g in grouping)
    elif isinstance(grouping, dict):
        return {k: map_grouping(fn, g) for k, g in grouping.items()}
    else:
        return fn(grouping)


def make_grouping_by_attr(schema, source, default=None):
    """
    Create a grouping from a schema by using the schema's scalar values to look up
    attributes in the provided source object.

    :param schema: A grouping of potential attributes in source
    :param source: Object to use to look up scalar grouping values from attributes
    :param default: Default value to use if attribute is not present in source
    :return: grouping
    """
    return map_grouping(lambda s: getattr(source, s, default), schema)


def make_grouping_by_key(schema, source, default=None):
    """
    Create a grouping from a schema by ujsing the schema's scalar values to look up
    items in the provided source object.

    :param schema: A grouping of potential keys in source
    :param source: Dict-like object to use to look up scalar grouping value using
        scalar grouping values as keys
    :param default: Default scalar value to use if grouping scalar key is not present
        in source
    :return: grouping
    """
    return map_grouping(lambda s: source.get(s, default), schema)


def make_schema_with_nones(grouping):
    """
    Create a grouping by replacing all grouping scalars values with None
    """
    return map_grouping(lambda _: None, grouping)


class SchemaValidationError(ValueError):
    def __str__(self):
        return "Grouping Schema Validation Error"


class SchemaTypeValidationError(SchemaValidationError):
    def __init__(self, value, full_schema, path, expected_type):
        self.value = value
        self.full_schema = full_schema
        self.path = list(path)
        self.expected_type = expected_type.__name__

    def __str__(self):
        return (
            "\n"
            f"    Schema: {self.full_schema}\n"
            f"    Path: {repr(self.path)}\n"
            f"    Expected type: {self.expected_type}\n"
            f"    Received value of type {type(self.value)}:\n"
            f"        {repr(self.value)}\n"
        )

    @classmethod
    def check(cls, value, full_schema, path, expected_type):
        if not isinstance(value, expected_type):
            raise SchemaTypeValidationError(value, full_schema, path, expected_type)


class SchemaLengthValidationError(SchemaValidationError):
    def __init__(self, value, full_schema, path, expected_len):
        self.value = value
        self.full_schema = full_schema
        self.path = list(path)
        self.expected_len = expected_len

    def __str__(self):
        return (
            "\n"
            f"    Schema: {self.full_schema}\n"
            f"    Path: {repr(self.path)}\n"
            f"    Expected length: {self.expected_len}\n"
            f"    Received value of length {len(self.value)}:\n"
            f"        {repr(self.value)}\n"
        )

    @classmethod
    def check(cls, value, full_schema, path, expected_len):
        if len(value) != expected_len:
            raise SchemaLengthValidationError(value, full_schema, path, expected_len)


class SchemaKeysValidationError(SchemaValidationError):
    def __init__(self, value, full_schema, path, expected_keys):
        self.value = value
        self.full_schema = full_schema
        self.path = list(path)
        self.expected_keys = expected_keys

    def __str__(self):
        return (
            "\n"
            f"    Schema: {self.full_schema}\n"
            f"    Path: {repr(self.path)}\n"
            f"    Expected keys: {self.expected_keys}\n"
            f"    Received value with keys {set(self.value.keys())}:\n"
            f"        {repr(self.value)}\n"
        )

    @classmethod
    def check(cls, value, full_schema, path, expected_keys):
        if set(value.keys()) != set(expected_keys):
            raise SchemaKeysValidationError(value, full_schema, path, expected_keys)


def validate_grouping(grouping, schema, full_schema=None, path=()):
    """
    Validate that the provided grouping conforms to the provided schema.
    If not, raise a SchemaValidationError
    """
    if full_schema is None:
        full_schema = schema

    if isinstance(schema, tuple):
        SchemaTypeValidationError.check(grouping, full_schema, path, tuple)
        SchemaLengthValidationError.check(grouping, full_schema, path, len(schema))

        for i, (g, s) in enumerate(zip(grouping, schema)):
            validate_grouping(g, s, full_schema=full_schema, path=path + (i,))
    elif isinstance(schema, dict):
        SchemaTypeValidationError.check(grouping, full_schema, path, dict)
        SchemaKeysValidationError.check(grouping, full_schema, path, set(schema))

        for k in schema:
            validate_grouping(
                grouping[k], schema[k], full_schema=full_schema, path=path + (k,)
            )
    else:
        pass
